import os
import unittest

import pandas as pd

from fast_form.config.configuration_loading import get_processing_config
from fast_form.outputting.ouput_data import output_data
from fast_form.outputting.process_document import process_document, process_document_and_add_to_validation_excel
from fast_form.outputting.utils_for_main import load_paths_for_processing_config, VALIDATION_EXCEL_NAME


class TestWholeProcess(unittest.TestCase):
    path_config_path = os.path.normpath(os.path.join(os.path.dirname(__file__),
                                                     os.path.normpath("form1_for_test/path_config.json")))
    paths_for_processing_config = load_paths_for_processing_config(path_config_path)
    document_path = os.path.join(paths_for_processing_config.folder_with_documents_path, "document.jpg")
    processing_config = get_processing_config(paths_for_processing_config)

    def test_letters_numbers_boxes_recognized_output_created(self):
        form_data = process_document(document_path=self.document_path, processing_config=self.processing_config)
        exp = pd.DataFrame(
            index=[0, 1, 2, 3],
            data={
                "name": ["Spali jste během Unihacku?", "Datum vyplnění", "Jméno týmu", "Chcete přepis, nebo zmrzlinu?"],
                "data": [3, "03 05 2020", "RYCHLY MQUCHY", 1]
            })

        df, images = output_data(form_data, "test_patient_id")
        pd.testing.assert_frame_equal(exp, df[["name", "data"]])

    def test_process_to_excel(self):
        validation_excel_path = os.path.join(self.paths_for_processing_config.folder_with_documents_path,
                                             VALIDATION_EXCEL_NAME)
        if os.path.exists(validation_excel_path):
            os.remove(validation_excel_path)
        process_document_and_add_to_validation_excel(document_path=self.document_path,
                                                     processing_config=self.processing_config,
                                                     excel_path=validation_excel_path)
        process_document_and_add_to_validation_excel(document_path=self.document_path,
                                                     processing_config=self.processing_config,
                                                     excel_path=validation_excel_path)



import logging
from typing import List

import numpy as np

from fast_form.field_recognizer.recognizing_dataclasses import RecognizingBoxResult, SingleChoiceStats, \
    RecognizingResult
from fast_form.structure_parser.form_structure_dataclasses import Field, FieldType

logger = logging.getLogger(__name__)

MULTIPLE_RESPONSES_FOUND = -1
NO_RESPONSE_FOUND = -2


def black_count_ratio(img: np.ndarray) -> int:
    ratio = sum(sum(img == 0)) / img.size
    assert 1 > ratio > 0
    return ratio


def recognize_single_choice_one_box(img: np.ndarray, single_choice_stats: SingleChoiceStats) -> RecognizingBoxResult:
    r = black_count_ratio(img)
    if r >= single_choice_stats.split_between:
        is_full = True
        dist = single_choice_stats.mean_full - single_choice_stats.split_between
        accuracy = np.clip(((r - single_choice_stats.split_between) - (single_choice_stats.mean_full - r)) / dist,
                           a_min=0,
                           a_max=1
                           )

    else:
        is_full = False
        dist = single_choice_stats.split_between - single_choice_stats.mean_empty
        accuracy = np.clip(((single_choice_stats.split_between - r) - (r - single_choice_stats.mean_empty)) / dist,
                           a_min=0,
                           a_max=1
                           )
    return RecognizingBoxResult(is_full, accuracy, img)


def recognize_single_choice(imgs: List[np.ndarray], single_choice_stats: SingleChoiceStats) -> RecognizingResult:
    recognizing_box_results = [recognize_single_choice_one_box(img, single_choice_stats) for img in imgs]
    accuracies_of_selected_choices = [(i, box_result.accuracy) for i, box_result in enumerate(recognizing_box_results)
                                      if box_result.recognized]
    if len(accuracies_of_selected_choices) > 1:
        logger.warning(f"Several possible selected choice boxes, returning {MULTIPLE_RESPONSES_FOUND}")
        selected = MULTIPLE_RESPONSES_FOUND
    elif len(accuracies_of_selected_choices) == 1:
        selected = accuracies_of_selected_choices[0][0]
    else:
        logger.warning(f"No possible selected choice boxes, returning {NO_RESPONSE_FOUND}")
        selected = NO_RESPONSE_FOUND
    return RecognizingResult(
        recognized=selected,
        recognizing_box_results=recognizing_box_results,
        accuracy=min(recognizing_box_results,
                     key=lambda recognizing_box_result: recognizing_box_result.accuracy).accuracy
    )


def calculate_box_stats(fields: List[Field]) -> SingleChoiceStats:
    black_ratios = [
        black_count_ratio(img) for field in fields for img in field.box_images if field.type == FieldType.SINGLE_CHOICE]

    average_blackness = np.mean(black_ratios)
    empty_mean_first_approximation = np.mean([box for box in black_ratios if box <= average_blackness])
    full_mean_first_approximation = np.mean([box for box in black_ratios if box >= average_blackness])
    split_between_first_approximation = (empty_mean_first_approximation + full_mean_first_approximation) / 2

    empty_mean = np.mean([box for box in black_ratios if box <= split_between_first_approximation])
    full_mean = np.mean([box for box in black_ratios if box >= split_between_first_approximation])
    split_between = (empty_mean + full_mean) / 2

    return SingleChoiceStats(mean_empty=empty_mean, mean_full=full_mean, split_between=split_between)



import math

import cv2
import numpy as np


def resize(img: np.ndarray, full_dimension: int, border: int):
    w, h = img.shape
    maxi = max([w, h])
    if maxi == 0:
        return np.array([])

    dimension = full_dimension - (border * 2)
    r = dimension / maxi
    new_w = round(w * r)
    new_h = round(h * r)

    res = cv2.resize(img, dsize=(new_h, new_w), interpolation=cv2.INTER_CUBIC)
    if max(res.shape) != dimension:
        raise ValueError(f"not good size of resized shape: {res.shape}, fix! ")

    bottom = math.ceil((dimension - new_w) / 2) + border
    top = math.floor((dimension - new_w) / 2) + border
    right = math.ceil((dimension - new_h) / 2) + border
    left = math.floor((dimension - new_h) / 2) + border
    res = cv2.copyMakeBorder(res, top, bottom, left, right, cv2.BORDER_CONSTANT, value=255)

    return res



from fast_form.config.configuration_dataclasses import Models
from fast_form.field_recognizer.recognize_boxes import calculate_box_stats, recognize_single_choice
from fast_form.field_recognizer.recognize_letters import recognize_chars
from fast_form.structure_parser.form_structure_dataclasses import FormPageData, FieldType


def recognize(form_data: FormPageData, models: Models) -> FormPageData:
    box_stats = calculate_box_stats(form_data.fields)
    fields = []
    for field in form_data.fields:
        if field.type == FieldType.LETTERS:
            field.recognizing_results = recognize_chars(field.box_images, models.model_letters,
                                                        models.letter_mapper)
        elif field.type == FieldType.NUMBERS:
            field.recognizing_results = recognize_chars(field.box_images, models.model_numbers,
                                                        models.number_mapper)
        elif field.type == FieldType.SINGLE_CHOICE:
            field.recognizing_results = recognize_single_choice(field.box_images, box_stats)
        else:
            raise ValueError(f"Unknown field type {field.type}")
        fields.append(field)
    form_data.fields = fields
    return form_data



from dataclasses import dataclass
from typing import List, Any, Dict

import cv2
import numpy as np

from fast_form.structure_parser.form_structure_dataclasses import FormStructure

ImageCv2 = np.ndarray


@dataclass
class ImageSiftResult:
    key_points: List[cv2.KeyPoint]
    descriptions: np.ndarray


@dataclass
class Template:
    sift: ImageSiftResult
    image: np.ndarray


@dataclass
class FormTemplates:
    templates: List[Template]


@dataclass
class Models:
    model_letters: Any
    model_numbers: Any
    letter_mapper: Dict
    number_mapper: Dict


@dataclass
class ProcessingConfig:
    models: Models
    templates: List[Template]
    form_structure: FormStructure


@dataclass
class PathsForProcessingConfig:
    template_path: str
    form_structure_path: str
    folder_with_documents_path: str
    final_excel_path: str = "result.xlsx"



import cv2

from fast_form.config.configuration_dataclasses import ImageCv2, ImageSiftResult


def get_image_sift_result(image: ImageCv2) -> ImageSiftResult:
    sift = cv2.SIFT_create()
    # find the keypoints and descriptors with SIFT
    key_points, descriptions = sift.detectAndCompute(image, None)
    return ImageSiftResult(key_points, descriptions)



import os
import unittest

from fast_form.config.configuration_loading import get_processing_config
from fast_form.outputting.process_document import process_document_and_add_to_validation_excel
from fast_form.outputting.utils_for_main import load_paths_for_processing_config, VALIDATION_EXCEL_NAME


class TestWholeProcess(unittest.TestCase):
    path_config_path = os.path.normpath(os.path.join(os.path.dirname(__file__),
                                                     os.path.normpath("form2_for_test/path_config.json")))
    paths_for_processing_config = load_paths_for_processing_config(path_config_path)
    document_path = os.path.join(paths_for_processing_config.folder_with_documents_path, "document.jpg")
    processing_config = get_processing_config(paths_for_processing_config)
    validation_excel_path = os.path.join(paths_for_processing_config.folder_with_documents_path,
                                         VALIDATION_EXCEL_NAME)

    def test_process_to_excel(self):
        if os.path.exists(self.validation_excel_path):
            os.remove(self.validation_excel_path)
        process_document_and_add_to_validation_excel(document_path=self.document_path,
                                                     processing_config=self.processing_config,
                                                     excel_path=self.validation_excel_path)



from typing import List

from fast_form.config.configuration_dataclasses import Template, ImageCv2
from fast_form.preprocessing.image_sift import get_image_sift_result
from fast_form.preprocessing.normalization import normalize
from fast_form.utils.image_loading import load_images_from_path

PDF_TEMPLATE = "template.pdf"
PICKLE_TEMPLATE = "template.pickle"
IMAGE_TEMPLATE = "template.png"
IMAGE_TEMPLATE_2 = "template.jpg"


def get_template_from_normalized_image(image: ImageCv2) -> Template:
    sift_result = get_image_sift_result(image)
    return Template(sift_result, image)


def get_templates(template_image_path: str) -> List[Template]:
    images = load_images_from_path(template_image_path)

    templates = [get_template_from_normalized_image(normalize(im)) for im in images]
    return templates



import cv2

from fast_form.field_recognizer.recognizing_dataclasses import RecognizingBoxResult, RecognizingResult
from fast_form.field_recognizer.resizing_for_recognizers import resize

EMNIST_SIZE = 28
LETTER_BORDER = 2
COLOR_RANGE = 255
NUMBER_OF_PIXELS_FOR_SPACE = 20


def is_not_space(img):
    if len(img) == 0:
        return False
    return sum(sum(img == 0)) > NUMBER_OF_PIXELS_FOR_SPACE


# not used anymore as it did not help performance
def thicken(img):
    new_img = cv2.bitwise_not(
        cv2.dilate(
            cv2.bitwise_not(img),
            cv2.getStructuringElement(cv2.MORPH_RECT, (3, 3)),
            iterations=1
        )
    )
    return new_img


def preprocess_img(img):
    return resize(
        img,
        EMNIST_SIZE,
        LETTER_BORDER
    )


def recognize_chars(imgs, model, result_mapper) -> RecognizingResult:
    recognizing_box_results = [recognize_char(preprocess_img(img), model, result_mapper) for img in imgs]
    return RecognizingResult(
        accuracy=min(recognizing_box_results,
                     key=lambda recognizing_box_result: recognizing_box_result.accuracy).accuracy,
        recognizing_box_results=recognizing_box_results,
        recognized="".join([res.recognized for res in recognizing_box_results])
    )


def recognize_char(img, model, result_mapper) -> RecognizingBoxResult:
    if is_not_space(img):
        img_p = img.copy()
        img_p = prepare_for_model_format(img_p)
        prediction, accuracy = predict(img_p, model, result_mapper)
        return RecognizingBoxResult(prediction, accuracy, img)
    else:
        return RecognizingBoxResult(" ", 1.0, img)


def predict(img, model, result_mapper):
    pred = model.predict(img)
    pred_val = result_mapper.get(pred[0].argmax())
    accuracy = pred[0].max()
    return pred_val, accuracy


def prepare_for_model_format(img):
    img = 1 - (img.reshape(1, EMNIST_SIZE, EMNIST_SIZE, 1).transpose() / COLOR_RANGE)
    return img



import argparse
import logging
import os

from fast_form.outputting.utils_for_main import load_paths_for_processing_config, process_to_validation_excel, \
    process_to_final_excel
from fast_form.version import VERSION

logging.basicConfig(
    format='%(asctime)s %(levelname)-8s %(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S')

logger = logging.getLogger(__name__)


def get_parser():
    """
    Creates a new argument parser.
    """
    par = argparse.ArgumentParser('fast-form')
    version = '%(prog)s ' + VERSION
    par.add_argument('--version', '-v', action='version', version=version)
    par.add_argument('--path_to_path_config',
                     type=str,
                     help="""Path to config file with paths to template pdf, leaving blank will run the file with
                      default configuration on path path_config.json. Required paths are:
                        {
                            "template_path": "path",
                            "form_structure_config_path": "path",
                            "folder_with_documents_path": "path",
                            "final_excel_path": "path"
                        }
                        """,
                     default=os.path.join(os.path.dirname(__file__), "tests", "form1_for_test", "path_config.json"))
    par.add_argument('--final_step',
                     action='store_true',
                     default=False,
                     help='If provided, the program will run the final step and output the final excel based on the'
                          'validation excel')
    return par


if __name__ == '__main__':
    logger.info("Starting app, loading tools needed for processing")
    parser = get_parser()
    args = parser.parse_args()
    paths_for_processing = load_paths_for_processing_config(args.path_to_path_config)
    if not args.final_step:
        process_to_validation_excel(paths_for_processing)
    else:
        process_to_final_excel(paths_for_processing)



import os
import tempfile
from typing import List, Tuple

import numpy as np
import openpyxl
import pandas as pd
from cv2 import cv2
from openpyxl import load_workbook

from fast_form.structure_parser.form_structure_dataclasses import FormStructure, FieldType

PATIENT_ID_FIELD = "patient_id"
SHEET_NAME = "questionnaire_data_output"


def output_data(form_data: FormStructure, maybe_patient_id: str) -> Tuple[pd.DataFrame, List[np.ndarray]]:
    form_dict = dict()
    form_images = []
    question_number = 0
    patient_id = maybe_patient_id
    for page_data in form_data.form_page_data.values():
        for field in page_data.fields:
            if field.name == PATIENT_ID_FIELD:
                assert isinstance(field.recognizing_results.recognized, str)
                read_patient_id = field.recognizing_results.recognized.replace(" ", "")
                if len(read_patient_id) > 0:
                    patient_id = read_patient_id

    for page_data in form_data.form_page_data.values():
        for field in page_data.fields:
            if field.name != PATIENT_ID_FIELD:
                field_dict = dict()
                field_dict["patient_id"] = patient_id
                field_dict['field_type'] = field.type.name
                field_dict['name'] = field.name
                field_dict['data'] = field.recognizing_results.recognized
                field_dict['acc'] = field.recognizing_results.accuracy
                form_images.append(field.img)
                form_dict[question_number] = field_dict
                question_number += 1

    df = pd.DataFrame.from_dict(form_dict, orient="index")
    return df, form_images


def save_data(df: pd.DataFrame, images: List[np.ndarray], excel_path: str):
    with tempfile.TemporaryDirectory() as temp_image_folder:

        if not os.path.exists(excel_path):
            df.to_excel(excel_path, index=False, sheet_name=SHEET_NAME)
            start_index = 1
            book = load_workbook(filename=excel_path)
            sheets_dict = {ws.title: ws for ws in book.worksheets}
        else:
            book = load_workbook(excel_path)
            writer = pd.ExcelWriter(excel_path, engine='openpyxl')
            writer.book = book
            sheets_dict = {ws.title: ws for ws in book.worksheets}
            writer.sheets = sheets_dict
            start_index = writer.sheets[SHEET_NAME].max_row
            df.to_excel(writer, sheet_name=SHEET_NAME, startrow=start_index, index=False, header=False)
            writer.save()
        sheet = sheets_dict[SHEET_NAME]
        for i, img, response, field_type in zip(df.index + start_index, images, df.data.values, df.field_type.values):
            if field_type != FieldType.SINGLE_CHOICE or response < 0:
                temp_image_path = os.path.join(temp_image_folder, f"{i}_img.png")
                height = 60
                width = int(height / img.shape[0] * img.shape[1])
                sheet.row_dimensions[i + 1].height = 50

                dim = (width, height)
                resized = cv2.resize(img, dim, interpolation=cv2.INTER_AREA)

                cv2.imwrite(temp_image_path, resized)
                img = openpyxl.drawing.image.Image(temp_image_path)
                img.anchor = f'F{i + 1}'
                sheet.add_image(img)
        book.save(excel_path)



import copy
from dataclasses import dataclass
from typing import List

import cv2
import numpy as np

from fast_form.structure_parser.form_structure_dataclasses import FormPageData, FieldType, Orientation, Field


@dataclass
class PageStructureParser:
    """
    Gets imdata in cv2 image
    """
    page_structure: FormPageData

    def process_page(self, form_img) -> FormPageData:
        fields = []
        page_structure = copy.deepcopy(self.page_structure)

        for field in page_structure.fields:
            field_crops = self.process_field(field, form_img)
            fields.append(field_crops)

        page_structure.fields = fields
        return page_structure

    def process_field(self, field_def: Field, form):

        x1 = field_def.top_left.x
        x2 = field_def.bottom_right.x
        y1 = field_def.top_left.y
        y2 = field_def.bottom_right.y

        field_def.img = form[y1:y2, x1:x2]
        if field_def.orientation == Orientation.VERTICAL:
            box_data = self.process_vertical_boxes(field_def)
        else:
            box_data = self.process_horizontal_boxes(field_def)

        field_def.box_images = box_data
        return field_def

    def process_horizontal_boxes(self, field_def: Field) -> List[np.ndarray]:
        box_data = []
        field_img = field_def.img
        x1 = field_def.top_left.x
        x2 = field_def.bottom_right.x
        space = field_def.space_between_boxes
        int_split = np.linspace(0, x2 - x1, field_def.number_of_boxes + 1)

        for xx1, xx2 in zip(int_split[:-1], int_split[1:]):
            box_img = field_img[:, int(xx1) + int(space / 2): int(xx2 - space / 2)]
            trimmed_box = self.trim_whitespace(box_img, field_def.type)
            box_data.append(trimmed_box)
        return box_data

    def process_vertical_boxes(self, field_def: Field) -> List[np.ndarray]:
        box_data = []
        field_img = field_def.img
        y1 = field_def.top_left.y
        y2 = field_def.bottom_right.y
        space = field_def.space_between_boxes
        int_split = np.linspace(0, y2 - y1 + space, field_def.number_of_boxes + 1)
        for yy1, yy2 in zip(int_split[:-1], int_split[1:]):
            box_img = field_img[int(yy1 - space / 2): int(yy2 - space / 2), :]

            trimmed_box = self.trim_whitespace(box_img, field_def.type)
            box_data.append(trimmed_box)

        return box_data

    @staticmethod
    def trim_whitespace(img: np.ndarray, field_type: FieldType) -> np.ndarray:
        if field_type is FieldType.SINGLE_CHOICE:
            return img

        gray = 255 * (img < 128).astype(np.uint8)
        coords = cv2.findNonZero(gray)  # Find all non-zero points (text)

        x, y, w, h = cv2.boundingRect(coords)  # Find minimum spanning bounding box
        trimmed_img = img[y:y + h, x:x + w]  # Crop the image
        return trimmed_img



import logging
from typing import List, Tuple

import cv2
import numpy as np

from fast_form.config.configuration_dataclasses import ImageCv2, ImageSiftResult, Template
from fast_form.preprocessing.image_sift import get_image_sift_result
from fast_form.preprocessing.normalization import normalize

ENOUGH_GOOD_MATCHES_FOR_FIT = 4

logger = logging.getLogger(__name__)


def get_good_matches_for_template(image_sift: ImageSiftResult, template_sift: ImageSiftResult) -> List[cv2.DMatch]:
    # based on https://towardsdatascience.com/image-stitching-using-opencv-817779c86a83

    bf = cv2.BFMatcher()
    matches = bf.knnMatch(image_sift.descriptions, template_sift.descriptions, k=2)

    # Apply ratio test
    good_matches = []
    for match in matches:
        if match[0].distance < 0.5 * match[1].distance:
            good_matches.append(match[0])
    return good_matches


def find_best_template_match(image_sift: ImageSiftResult,
                             templates: List[Template]) -> Tuple[List[cv2.DMatch], Template, int]:
    all_good_matches = [(get_good_matches_for_template(image_sift, template.sift), template, i)
                        for i, template in enumerate(templates)]
    return max(all_good_matches, key=lambda good_matches_template: len(good_matches_template[0]))


def fit_image_to_templates(image: ImageCv2, templates: List[Template]) -> Tuple[ImageCv2, int]:
    image_sift = get_image_sift_result(image)
    good_matches, template, template_index = find_best_template_match(image_sift, templates)

    if len(good_matches) >= ENOUGH_GOOD_MATCHES_FOR_FIT:
        image_similarities = np.float32(
            [image_sift.key_points[match.queryIdx].pt for match in good_matches]
        ).reshape(-1, 1, 2)
        template_similarities = np.float32(
            [template.sift.key_points[match.trainIdx].pt for match in good_matches]).reshape(-1, 1, 2)
        homography_matrix, _ = cv2.findHomography(image_similarities, template_similarities, cv2.RANSAC, 5.0)

        return cv2.warpPerspective(image, homography_matrix,
                                   (template.image.shape[1], template.image.shape[0])), template_index
    else:
        raise AssertionError("Cant find enough keypoints.")


def preprocess(im: np.ndarray, form_templates: List[Template]) -> np.ndarray:
    """Runs the full pipeline:

    - Loads input image
    - Normalizes image
    - fits to original
    """
    normalized_image = normalize(im)

    fitted_image, template_index = fit_image_to_templates(normalized_image, templates=form_templates)

    return fitted_image



import cv2
import numpy as np


def normalize(img: np.ndarray) -> np.ndarray:
    """Converts `im` to black and white.

    Applying a threshold to a grayscale image will make every pixel either
    fully black or fully white."""

    im_blur = cv2.GaussianBlur(img, (7, 7), 0)

    im_gray = cv2.cvtColor(im_blur, cv2.COLOR_BGR2GRAY)
    im_thresh = cv2.adaptiveThreshold(im_gray, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 11, 2)
    return im_thresh



from typing import List

import numpy as np
from PIL.PpmImagePlugin import PpmImageFile
from cv2 import cv2
from pdf2image import convert_from_path

from fast_form.config.configuration_dataclasses import ImageCv2


def to_cv_image(pid_image: PpmImageFile) -> ImageCv2:
    return np.array(pid_image.convert('RGB'))


def load_images_from_path(template_image_path: str) -> List[np.array]:
    ending = template_image_path.lower().split(".")[-1]

    if ending == "pdf":
        images = [to_cv_image(im_pip) for im_pip in
                  convert_from_path(template_image_path, dpi=300)]
    elif ending in {"jpg", "png", "jpeg"}:
        images = [cv2.imread(template_image_path)]
    else:
        raise ValueError(f"Uknown file format file: {template_image_path}")
    return images



from dataclasses import dataclass
from enum import Enum
from typing import Optional, List, Dict

import numpy as np

from fast_form.field_recognizer.recognizing_dataclasses import RecognizingResult


@dataclass
class Point:
    x: int
    y: int


class Orientation(str, Enum):
    HORIZONTAL = 'HORIZONTAL'
    VERTICAL = 'VERTICAL'


class FieldType(str, Enum):
    LETTERS = 'LETTERS'
    NUMBERS = 'NUMBERS'
    SINGLE_CHOICE = 'SINGLE_CHOICE'


@dataclass
class Field:
    orientation: Orientation
    space_between_boxes: int
    number_of_boxes: int
    name: str
    type: FieldType
    top_left: Point
    bottom_right: Point
    img: Optional[np.ndarray] = None
    recognizing_results: Optional[RecognizingResult] = None
    box_images: Optional[List[np.ndarray]] = None


@dataclass
class FormPageData:
    fields: List[Field]


@dataclass
class FormStructure:
    form_page_data: Dict[str, FormPageData]
    page_count: int



from typing import List

from fast_form.structure_parser.form_structure_dataclasses import Field, Orientation, FieldType, Point


def get_fields_for_page(
        questions_per_page: int,
        y_start: int,
        x_start: int,
        width: int,
        line_step: int,
        line_height: int,
        first_question_number: int) -> List[Field]:
    fields = []
    question_number = first_question_number
    for i in range(questions_per_page):
        fields.append(
            Field(
                orientation=Orientation.HORIZONTAL,
                space_between_boxes=80,
                name=f"{question_number}",
                type=FieldType.SINGLE_CHOICE,
                top_left=Point(x_start, y_start + i * line_step),
                bottom_right=Point(x_start + width, y_start + line_height + i * line_step),
                number_of_boxes=4
            )
        )
        question_number += 1
    return fields



# coding=utf-8

"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# To use a consistent encoding
from codecs import open
from os import path

# Always prefer setuptools over distutils
from setuptools import find_packages, setup

from fast_form.version import VERSION

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md')) as f:
    long_description = f.read()

setup(
    name='fast-form',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=VERSION,

    description='Automatically process scanned forms',
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    # The project's main homepage.
    url='https://github.com/kubantjan/fast-form',
    # Author details
    author='Mild Blue',
    author_email='info@mild.blue',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    # What does your project relate to?
    keywords='form questionnaire omr',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'numpy>=1.18.3',
        'opencv-contrib-python>=4.4.0.42',
        'pandas>=1.1.4',
        'tensorflow<=2.1.0',
        'keras<=2.3.1',
        'autopep8>=1.5.4',
        'dacite>=1.5.1',
        'openpyxl>=3.0.5',
        'pdf2image>=1.14.0',
        'xlrd>=1.2.0',
        'h5py < 3.0.0'
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': [
            'matplotlib >= 3.3.1',
            'pre-commit >= 2.6.0',
            'jupyter',
            'ipykernel >= 5.3.4'
        ]
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
        'sample': [],
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={},
)



from dataclasses import dataclass
from typing import Union, List

import numpy as np


@dataclass
class RecognizingBoxResult:
    recognized: Union[str, bool, int]
    accuracy: float
    img_transf: np.ndarray


@dataclass
class RecognizingResult:
    recognized: Union[int, bool]
    recognizing_box_results: List[RecognizingBoxResult]
    accuracy: float


@dataclass
class SingleChoiceStats:
    split_between: float
    mean_empty: float
    mean_full: float



import json
import logging
import os

import dacite
import pandas as pd
from openpyxl import load_workbook

from fast_form.config.configuration_dataclasses import PathsForProcessingConfig
from fast_form.config.configuration_loading import get_processing_config
from fast_form.outputting.process_document import process_document_and_add_to_validation_excel

SHEET_WITH_RESULTS = "automatic_results"
VALIDATION_EXCEL_NAME = "validation_excel.xlsx"

logger = logging.getLogger(__name__)


def process_to_validation_excel(paths_for_processing_config: PathsForProcessingConfig):
    processing_config = get_processing_config(paths_for_processing_config)
    validation_excel_path = os.path.join(paths_for_processing_config.folder_with_documents_path, VALIDATION_EXCEL_NAME)
    if os.path.exists(validation_excel_path):
        os.remove(validation_excel_path)

    document_names = [file for file in os.listdir(paths_for_processing_config.folder_with_documents_path) if
                      file.endswith('.pdf') or file.endswith(".jpg") or file.endswith(".png")]
    logging.info(f"Processing to validation excel from {paths_for_processing_config.folder_with_documents_path}. Number"
                 f" of documents is {len(document_names)}")
    for document_name in document_names:
        process_document_and_add_to_validation_excel(
            os.path.join(paths_for_processing_config.folder_with_documents_path, document_name),
            processing_config,
            validation_excel_path
        )
    logger.info(
        f"Successfully processed all documents from '{paths_for_processing_config.folder_with_documents_path}' to "
        f"validation excel '{validation_excel_path}'")


def process_to_final_excel(paths_for_processing_config: PathsForProcessingConfig):
    validation_excel_path = os.path.join(paths_for_processing_config.folder_with_documents_path,
                                         VALIDATION_EXCEL_NAME)
    if not os.path.exists(validation_excel_path):
        raise FileNotFoundError(f"Validation excel with path '{validation_excel_path}' must exist")

    validation_df = pd.read_excel(validation_excel_path)
    one_patient_per_row_df = (
        validation_df
            .set_index(['patient_id', 'name'])
            .data
            .apply(lambda response_or_error: response_or_error if response_or_error >= 0 else "")
            .unstack(level=1)
    )

    combined_one_patient_per_row_df = one_patient_per_row_df

    if os.path.exists(paths_for_processing_config.final_excel_path):
        xl = pd.ExcelFile(paths_for_processing_config.final_excel_path)
        if SHEET_WITH_RESULTS in xl.sheet_names:  # see all sheet names
            old_one_patient_per_row_df = xl.parse(
                index_col="patient_id",
                keep_default_na=False,
                na_values=[],
                sheet_name=SHEET_WITH_RESULTS)

            combined_one_patient_per_row_df = one_patient_per_row_df.combine_first(old_one_patient_per_row_df)

    book = load_workbook(paths_for_processing_config.final_excel_path)
    writer = pd.ExcelWriter(paths_for_processing_config.final_excel_path, engine='openpyxl')
    if SHEET_WITH_RESULTS in book.sheetnames:
        book.remove(book[SHEET_WITH_RESULTS])
    writer.book = book
    combined_one_patient_per_row_df.to_excel(writer, sheet_name=SHEET_WITH_RESULTS)
    writer.save()
    logger.info(f"Successfully processed from validation excel: '{validation_excel_path}'"
                f" to final excel '{paths_for_processing_config.final_excel_path}'")


def load_paths_for_processing_config(path_configuration: str) -> PathsForProcessingConfig:
    def join(rel_path):
        return os.path.join(os.path.dirname(path_configuration), rel_path)

    with open(path_configuration, "rb") as f:
        relative_paths = dacite.from_dict(data_class=PathsForProcessingConfig, data=json.loads(f.read()))
    return PathsForProcessingConfig(
        template_path=join(relative_paths.template_path),
        folder_with_documents_path=join(relative_paths.folder_with_documents_path),
        final_excel_path=join(relative_paths.final_excel_path),
        form_structure_path=join(relative_paths.form_structure_path)
    )



import json
import logging
import os

from keras.models import model_from_json

logger = logging.getLogger(__name__)


def load_model(model_structure_path, model_weights_path):
    # load json and create model
    # json_file = open('model_data/model.json', 'r')

    # load model structure
    with open(model_structure_path, 'r') as json_file:
        loaded_model_json = json_file.read()
    loaded_model = model_from_json(loaded_model_json)
    # load weights into new model
    # loaded_model.load_weights("model_data/model.h5")

    # load model weights
    loaded_model.load_weights(model_weights_path)
    logger.info(f"Loaded model {os.path.basename(model_structure_path)} from disk")

    # evaluate loaded model on test data
    loaded_model.compile(loss='categorical_crossentropy',  # using the cross-entropy loss function
                         optimizer='adam',  # using the Adam optimiser
                         metrics=['accuracy'])  # reporting the accuracy

    return loaded_model


# scores = model.evaluate(x_test, y_test, verbose=0)
# print("%s: %.2f%%" % (loaded_model.metrics_names[1], scores[1] * 100))

def load_result_mapper(path):
    with open(path, 'r') as f:
        raw = json.load(f)
    return {int(k): v for k, v in raw.items()}



import json
import os

import dacite
from dacite import Config

from fast_form.config.configuration_dataclasses import Models, PathsForProcessingConfig, ProcessingConfig
from fast_form.field_recognizer.model import load_model
from fast_form.preprocessing.templating import get_templates
from fast_form.structure_parser.form_structure_dataclasses import FormStructure, FieldType, Orientation

LETTERS_MODEL_JSON = "model_letters.json"
LETTERS_MODEL_H5 = "model_letters.h5"
NUMBERS_MODEL_JSON = "model_numbers.json"
NUMBERS_MODEL_H5 = "model_numbers.h5"
LETTER_MAPPER_JSON = "model_letters_mapping.json"
NUMBER_MAPPER_JSON = "model_numbers_mapping.json"


def load_models():
    model_data_location = os.path.join(os.path.dirname(__file__), "..", "model_data")
    model_letters_structure_path = f"{model_data_location}/{LETTERS_MODEL_JSON}"
    model_letters_weights_path = f"{model_data_location}/{LETTERS_MODEL_H5}"
    model_numbers_structure_path = f"{model_data_location}/{NUMBERS_MODEL_JSON}"
    model_numbers_weights_path = f"{model_data_location}/{NUMBERS_MODEL_H5}"
    letter_mapper_path = f"{model_data_location}/{LETTER_MAPPER_JSON}"
    number_mapper_path = f"{model_data_location}/{NUMBER_MAPPER_JSON}"

    model_letters = load_model(model_letters_structure_path, model_letters_weights_path)
    model_numbers = load_model(model_numbers_structure_path, model_numbers_weights_path)

    # load result mapper

    with open(letter_mapper_path, "r") as json_file:
        letter_mapper = json.load(json_file)
        letter_mapper = {int(k): v for k, v in letter_mapper.items()}

    with open(number_mapper_path, "r") as json_file:
        number_mapper = json.load(json_file)
        number_mapper = {int(k): v for k, v in number_mapper.items()}

    return Models(model_letters=model_letters, model_numbers=model_numbers, letter_mapper=letter_mapper,
                  number_mapper=number_mapper)


def load_form_structure(form_structure_config_path: str) -> FormStructure:
    with open(form_structure_config_path, 'r') as f:
        return dacite.from_dict(data_class=FormStructure, data=json.load(f),
                                config=Config(cast=[FieldType, Orientation]))


def get_processing_config(paths: PathsForProcessingConfig) -> ProcessingConfig:
    return ProcessingConfig(
        models=load_models(),
        templates=get_templates(paths.template_path),
        form_structure=load_form_structure(paths.form_structure_path),
    )



import os

from fast_form.config.configuration_dataclasses import ProcessingConfig
from fast_form.field_recognizer.recognize_all import recognize
from fast_form.outputting.ouput_data import output_data, save_data
from fast_form.preprocessing.preprocess import preprocess
from fast_form.structure_parser.form_structure_dataclasses import FormStructure
from fast_form.structure_parser.page_structure_parser import PageStructureParser
from fast_form.utils.image_loading import load_images_from_path


def process_document_and_add_to_validation_excel(document_path: str, processing_config: ProcessingConfig,
                                                 excel_path: str):
    if os.path.exists(document_path):
        maybe_patient_id = os.path.basename(document_path).split(".")[0]
        form_data = process_document(processing_config, document_path=document_path)
        df, images = output_data(form_data, maybe_patient_id=maybe_patient_id)

        save_data(df, images, excel_path=excel_path)
    else:
        raise FileExistsError("Document file not found")


def process_document(processing_config: ProcessingConfig,
                     document_path: str) -> FormStructure:
    images = load_images_from_path(document_path)

    assert len(processing_config.templates) == processing_config.form_structure.page_count
    assert len(images) == processing_config.form_structure.page_count, f"len images is {len(images)} and page count " \
                                                                       f"is " \
                                                                       f"{processing_config.form_structure.page_count}"

    for image, template, (page_name, page_structure) in zip(images,
                                                            processing_config.templates,
                                                            processing_config.form_structure.form_page_data.items()):
        image = preprocess(image, [template])
        page_structure_parser = PageStructureParser(page_structure)
        page_data = page_structure_parser.process_page(image)
        page_data = recognize(page_data, processing_config.models)
        processing_config.form_structure.form_page_data[page_name] = page_data

    return processing_config.form_structure
